#ifndef MARISA_GRIMOIRE_TRIE_H_
#define MARISA_GRIMOIRE_TRIE_H_

#include "trie/state.h"
#include "trie/louds-trie.h"

namespace marisa {
namespace grimoire {

using trie::State;
using trie::LoudsTrie;

}  // namespace grimoire
}  // namespace marisa

#endif  // MARISA_GRIMOIRE_TRIE_H_
