#ifndef MARISA_GRIMOIRE_IO_H_
#define MARISA_GRIMOIRE_IO_H_

#include "io/mapper.h"
#include "io/reader.h"
#include "io/writer.h"

namespace marisa {
namespace grimoire {

using io::Mapper;
using io::Reader;
using io::Writer;

}  // namespace grimoire
}  // namespace marisa

#endif  // MARISA_GRIMOIRE_IO_H_
